(*****************************************************************************
 NICODEMOS version 1.1 (Aug 2012)
 Author: Ayres Freitas (afreitas@pitt.edu)

 Release notes:
 
 NICODEMOS may be used freely for non-commercial purposes, with proper
 acknowledgement of the author and citation of the original publication:
  A. Freitas, arXiv:1205.3515
 If you include it as part of a different program, proper acknowledgement 
 must be given in the documentation of the program.
 
 The NICODEMOS package includes the package FORMAT 2.0 by Mark Sofroniou
 (Copyright 1992-2003,  Mark Sofroniou).
 
 It furthermore requires the CUBA numerical integration library by Thomas Hahn, 
 which can be obtained from http://www.feynarts.de/cuba/Cuba-1.4.tar.gz.

 The author does not promise that this software works, but bug reports are
 welcome.

 For further information see the manual.
*****************************************************************************
 Version history:
 ----------------
 
 version 1.1 (Aug 2012): added capability for treating two-loop integrals
                         with UV and subloop IR singularities
			 (i.e. functionality described in arXiv:1205.3515)
 
 version 1.0 (May 2012): first version, includes only one-loop functionality
 
*****************************************************************************)

$NicoDir = DirectoryName[File /. FileInformation[$Input]];
Get[ToFileName[$NicoDir, "paths"]];

Get[ToFileName[$NicoDir, "Format.m"]];
Off[AssignFunction::undef];

Get[ToFileName[$NicoDir, "onecode.m"]];
Get[ToFileName[$NicoDir, "tusub.m"]];
Get[ToFileName[$NicoDir, "ttools.m"]];


TwoLoop::resv = "ieps,jg,jgm,li2,lim,lip,xlog,x,y,z,$q,$q1,$q2 are reserved symbols.";
TwoLoop::ldef = "The objects 'pars', 'momext', 'rels' and/or 'muUV2' are undefined.";
TwoLoop::subdiv = "Cannot handle subloop UV divergences in both subloops.\n" <>
                  "See arXiv:1205.3515 for more information.";
TwoLoop::divbox = "The seems to be a UV-divergent subloop box or higher" <>
                  " n-point function, which is not expected in renormalizable" <>
		  " theories.";


SetOptions[WriteCode, SimpFunc->Identity];

QSort1[exprin_] := Sort[exprin, #2=!=$q1 && (#1===$q1 || OrderedQ[{#1,#2}]) &];
TSort1[exprin_] := Sort[exprin, #2=!=$q1.$q1 && (#1===$q1.$q1 || OrderedQ[{#1,#2}]) &];
QSort2[exprin_] := Sort[exprin, #2=!=$q2 && (#1===$q2 || OrderedQ[{#1,#2}]) &];
TSort2[exprin_] := Sort[exprin, #2=!=$q2.$q2 && (#1===$q2.$q2 || OrderedQ[{#1,#2}]) &];


$tl$tim = 0;
PrintStatus[step_Integer, vlevel_Integer] := Module[{},
  If[step > 1 && vlevel > 1,
    Print[" (time used: ", Round[TimeUsed[]-$tl$tim, 0.1], "s)"]
  ];
  $tl$tim = TimeUsed[];
  If[vlevel > 0 && step < 5, Print["Step ", step, " of 4"]];
];


Options[TwoLoop] = {SimpLevel -> 3,
                    StatusInfo -> 0};

TwoLoop[aden_, anum_, q1_, q2_, opts___Rule] := 
Module[{simplev,num,den,lprop,k12div,k1div,k2div,den12,den12a,den12b,sub12,
        i,j,k,n,n1,den1,exp1,sum1,trn1,jac1,int1,den2,exp2,sum2,trn2,jac2,int2,
	osub1,osuba,osubb,deni,k2shift,abbc,vars,nvar,varr,tures,res2,old2},
  simplev = SimpLevel /. {opts} /. Options[TwoLoop];
  stainf = StatusInfo /. {opts} /. Options[TwoLoop];
  If[!FreeQ[{aden,anum,q,varc}, ieps|jg|jgm|li2|lim|lip|xlog|x|y|z|$q|$q1|$q2],
    Message[TwoLoop::resv]; Return[]];
  If[!ValueQ[pars] || !ValueQ[momext] || !ValueQ[rels] || !ValueQ[muUV2],
    Message[TwoLoop::ldef]; Return[]];
  $ol$lsubs = tures = {};
  den = aden /. {q1->$q1, q2->$q2};
  num = anum /. {q1->$q1, q2->$q2};
  $ol$subt = $ol$subt /. {q1->$q1, q2->$q2};
  
  PrintStatus[1, stainf];

  (* perform one-loop integration for IR subtraction terms: *)
  k = 0;
  For[i = Length[$ol$subt], i > 0, i--,
    If[!FreeQ[$ol$subt[[i]], $q2], $ol$subt = $ol$subt /. $q2->$q1];
    If[!FreeQ[$ol$subt[[i]], $q1],
      osub1 = List @@ Together[$ol$subr /. $ol$subt[[i]]];
      osubb = Cases[osub1, ($q1.$q1 + (_:0))^(n_/;n<0)];
      osubb = 1/osubb //. {arga___, argb_^n_Integer, argc___} :> 
      			   {arga, argb^(n-1), argb, argc} /; n>1;
      osuba = Times @@ DeleteCases[osub1, ($q1.$q1 + (_:0))^(n_/;n<0)];
      osubb = OneLoop[osubb, Select[osuba, !FreeQ[#,$q1]&], $q1, 
                      SimpLevel -> simplev, NestedExpressions -> False];
      osuba = Select[osuba, FreeQ[#,$q1]&];
      k++;
      If[Coefficient[osuba, $e, -2] === 0,
	$ol$subt = ReplacePart[$ol$subt, i -> 
                        	{xe[k] -> Coefficient[osuba, $e, -1],
				 xf[k] -> Coefficient[osuba, $e, 0],
				 xd[k] -> Coefficient[osuba, $e, 1]}];
	osubb[[-1,-1]] *= (xe[k]/$e + xf[k] + xd[k]*$e),
      (* else *)
	$ol$subt = ReplacePart[$ol$subt, i -> 
                        	{xee[k] -> Coefficient[osuba, $e, -2],
				 xe[k] -> Coefficient[osuba, $e, -1],
				 xf[k] -> Coefficient[osuba, $e, 0],
				 xd[k] -> Coefficient[osuba, $e, 1]}];
	osubb[[-1,-1]] *= (xee[k]/$e^2 + xe[k]/$e + xf[k] + xd[k]*$e);
      ];
      If[FreeQ[osubb, x[_]],
        $ol$subt = Insert[$ol$subt, (osubb /. $ol$res->$ol$subr /. ieps->0), i+1],
        tures = Append[tures, osubb]
      ]
    ]
  ];
  $ol$subt = Flatten[$ol$subt];
  $ol$moms = Alternatives @@ Join[momext, {$q1,$q2}];

  (* take care of propagators left behind in 'num' from subtractions: *)
  lprop = 1/(($q1+___).($q1+___)+(_:0)) | 1/(($q2+___).($q2+___)+(_:0)) |
          1/((_:0)-($q1+___).($q1+___)) | 1/((_:0)-($q2+___).($q2+___));
  If[!FreeQ[num, lprop],
    num = Together[num];
    den = Join[den, Map[1/# &, Cases[num, lprop]]];
    num = DeleteCases[num, lprop];
    For[i = 1, i <= Length[den], i++,
      If[MatchQ[den[[i]], (_:0)-Dot[_,_]],
        den[[i]] = -den[[i]];
	num = -num;
      ]
    ]
  ];

  (* evaluate global UV singularity: *)
  While[True,
    (* power-counting of singularity: *)
    k12div = (Length[CoefficientList[num /. Dot->Times /. $q2->$q1, $q1]]-1)/2 - 
    	      Length[den];
    If[k12div < -4, Break[]];	(* not UV-singular (anymore) -> exit loop *)
    (* construct subtraction term by setting external momenta to zero: *)
    den12 = den /. Map[#->0 &, momext];
    If[FreeQ[den12, $q1+$q2],
      sub12 = TRed[den12, DotSimplify[num], $q1, $q2] /. rels,
      sub12 = TRed[den12 /. $q1+$q2 -> $q1-$q2, 
      		   DotSimplify[num /. $q2->-$q2], $q1, $q2] /. rels
    ];
    $ol$subt = Join[$ol$subt, {$ol$subr -> sub12}];
    (* put together subtracted integrand: *)
    For[den12a = den12b = {}; i = 1, i <= Length[den12], i++,
      If[den12[[i]] === den[[i]], 
	den12b = Append[den12b, den12[[i]]],
	den12a = Append[den12a, den12[[i]]]
      ]
    ];
    num = num*((Times @@ den12a) - 
               (Times @@ den / (Times @@ den12b)));
    den = Join[den, den12a];
  ];
  (* determine subloop divergences: *)
  k1div = (Length[CoefficientList[num /. Dot->Times, $q1]]-1)/2 - 
    	    Length[Select[den, !FreeQ[#,$q1]&]];
  k2div = (Length[CoefficientList[num /. Dot->Times, $q2]]-1)/2 - 
    	    Length[Select[den, !FreeQ[#,$q2]&]];
  (* make sure that UV divergences do not exist in both subloops: *)
  If[k1div >= -2 && k2div >= -2,
    Message[TwoLoop::subdiv]; Return[]];
  (* shuffle the UV subloop divergence into the $q1 loop, if necessary: *)
  If[k2div >= -2,
    den = den /. $q1->$q /. $q2->$q1 /. $q->$q2;
    num = num /. $q1->$q /. $q2->$q1 /. $q->$q2;
  ];
  
  (* work on $q1 subloop: *)
  den1 = Sort[Union[Select[den, !FreeQ[#,$q1]&]], 
	  OrderedQ[{#2,#1} /. Dot[_,_]->0]&];
  exp1 = Map[Count[den, #]&, den1];	(* power of each denominator term *)
  n = 0;
  (* introduce Feynman parameters: *)
  den1 = Plus @@ Map[z[++n]*# &, den1];
  sum1 = Plus @@ Table[z[j], {j,n}];
  (* variable transformation onto hyper-cube: *)
  trn1 = Table[z[j] -> Product[y[k], {k,j}]*If[j===n, 1, 1-y[j+1]], {j,n}];
  jac1 = Det[Outer[D, Map[Last,trn1], Table[y[j],{j,n}]]];
  trn1 = trn1 /. y[1]->1;
  jac1 = jac1 /. y[1]->1;
  den1 = Collect[DotSimplify[den1], Dot[_,_], Simplify[#, sum1==1]&];
  If[simplev > 0,
    num = Collect[DotSimplify[num] /. rels, Dot[_,_], Simplify]
  ];
  (* shift $q1 to eliminate linear term: *)
  num = num /. $q1 -> $q1 - Coefficient[den1 /. Dot->Times, $q1, 1]/2;
  den1 = den1 /. $q1 -> $q1 - Coefficient[den1 /. Dot->Times, $q1, 1]/2;
  den1 = Collect[DotSimplify[den1], $q1.$q1, Simplify[#, sum1==1]&];
  num = DotSimplify[num] /. rels;
  If[simplev > 1,
    num = Collect[num, Dot[_,_], Simplify],
    num = Collect[num, Dot[_,_], TurboSimplify]
  ];
  dens = den1^(-Plus @@ exp1) * Gamma[Sum[exp1[[j]], {j,n}]] * 
	  (Product[z[j]^(exp1[[j]]-1)/Gamma[exp1[[j]]], {j,n}]);
  (* make sure we have always the same sign convention: *)
  dens = dens /. (A_ - $q1.$q1)^expd_ :> (-1)^expd * ($q1.$q1 - A)^expd;
  num *= Collect[Numerator[dens] /. rels, Dot[_,_], Simplify];
  dens = 1/Denominator[dens] /. rels;
  PrintStatus[2, stainf];
  (* collect all momentum dot product in numerator that involve the loop
     momenta and give them the header T: *)
  num = Expand[num, Dot[_,__]];
  num = num /. Power[Dot[arga_,argb_], nA_:1] :> 
		  (T @@ Table[Dot[arga,argb], {nA}]) /; 
		  !FreeQ[Dot[arga,argb], $q1|$q2] //.
               T[arga__] * T[argb__] :> T[arga,argb];
  (* tensors of odd rank in $q1 are zero;
     tensors of even rank in $q1 evaluate to combinations of metric tensors: *)
  num = num /. expt_T :> 0 /; OddQ[Length[Position[expt, $q1]]];
  num = num /. expd_Dot :> QSort1[expd] /. expt_T :> TSort1[expt];
  num = num  //. 
       T[argt1___, $q1.arga_, $q1.argb_, $q1.argc_, $q1.argd_, $q1.arge_, $q1.argf_,
	 argt2___] :>
              1/($d*(2+$d)*(4+$d)) * T[argt1, $q1.$q1, $q1.$q1, $q1.$q1, argt2] * 
		(arga.argb*argc.argd*arge.argf + arga.argc*argb.argd*arge.argf +
		 arga.argd*argb.argc*arge.argf + arga.argb*argc.arge*argd.argf +
		 arga.argc*argb.arge*argd.argf + arga.arge*argb.argc*argd.argf +
		 arga.argb*argc.argf*argd.arge + arga.argc*argb.argf*argd.arge +
		 arga.argf*argb.argc*argd.arge + arga.argd*argb.arge*argc.argf +
		 arga.arge*argb.argd*argc.argf + arga.argd*argb.argf*argc.arge +
		 arga.argf*argb.argd*argc.arge + arga.arge*argb.argf*argc.argd +
		 arga.argf*argb.arge*argc.argd)  /;
		FreeQ[{argt1,argt2}, $q1.(mom_/;mom=!=$q1)] && 
		FreeQ[{arga,argb,argc,argd,arge,argf}, $q1]  //.
       T[argt1___, $q1.arga_, $q1.argb_, $q1.argc_, $q1.argd_, argt2___] :>
              1/($d*(2+$d)) * T[argt1, $q1.$q1, $q1.$q1, argt2] * 
	      (arga.argb*argc.argd + arga.argc*argb.argd + arga.argd*argb.argc) /;
		FreeQ[{argt1,argt2}, $q1.(mom_/;mom=!=$q1)] && 
		FreeQ[{arga,argb,argc,argd}, $q1]  //.
       T[argt1___, $q1.arga_, $q1.argb_, argt2___] :> 
	      1/$d*T[argt1, $q1.$q1, argt2] * arga.argb /;
	     	  FreeQ[{argt1,argt2}, $q1.(mom_/;mom=!=$q1)] && 
		  FreeQ[{arga,argb}, $q1];
  num = num /. expt_T :> TSort1[expt];
  If[simplev > 2,
    num = Collect[num //. rels, {T[__], Dot[_,_]}, TurboSimplify],
    num = Collect[num //. rels, {T[__], Dot[_,_]}]
  ];
  int1 = num * dens;
  (* cancel ($q1)^2 in numerator against denominator: *)
  int1 = int1 //. T[$q1.$q1, argt___] :> $q1^2 * T[argt] /.
		 T[]->1 /. Dot[$q1,$q1]->$q1^2;
  int1 = int1 //. ((R1_:1)*$q1^na_ + (R2_:0)) * ($q1^2 + A_)^nA_ :> 
          R1*($q1^2 + A)^(nA+na/2) + 
	  Collect[R2 - R1*(($q1^2+A)^(na/2) - $q1^na), {$q1,Dot[_,_]}]*
	    ($q1^2 + A)^nA /; FreeQ[R1, $q1];
  (* this cancellation may leave $q2-dependent terms outside of the T symbols;
     now make sure they are all moved inside the T symbols: *)
  int1 = Collect[int1, {($q1^2 + _)^_, T[__]}, 
		  # /. Power[Dot[arga_,argb_], nA_:1] :> 
			  (T @@ Table[Dot[arga,argb], {nA}]) /; 
			  !FreeQ[Dot[arga,argb], $q2] &];
  int1 = Expand[int1, T[__]];
  int1 = int1 //. {T[arga__] * T[argb__] :> T[arga,argb],
		   T[arga__]^2 :> T[arga,arga]};
  int1 = int1 /. expd_Dot :> QSort2[expd] /. expt_T :> TSort2[expt];
  If[simplev > 2,
    int1 = Collect[int1, {($q1^2 + _)^_, T[__]}, Simplify],
    int1 = Collect[int1, {($q1^2 + _)^_, T[__]}]
  ];
  (* UV subloop subtraction term: *)  
  osub1 = (Coefficient[int1 /. ($q1^2 + A_)^(-2) -> $tl$subdiv, $tl$subdiv] /. 
	   trn1 /. z[nn_] :> y[nn]) * jac1;
  If[osub1 =!= 0,
    vars = Union[Extract[osub1, Position[osub1, y[_]]]];
    If[vars === {},
      osuba = Simplify[osub1],
    (* else *)
      If[Length[vars] > 2, Message[TwoLoop::divbox]; Return[]];
      osub1 = Collect[osub1, T[__], Simplify];
      osuba = Integrate @@ Prepend[Map[{#,0,1}&, vars], osub1];
      If[simplev > 0,  osuba = Simplify[osuba, $d>3]];
    ];  
    (* perform one-loop integration for subtraction term: *)
    osubb = OneLoop[Select[den, FreeQ[#,$q1]&], 
	      Gamma[2-$d/2]*muUV2^($d/2-2) * osuba /. T->Times, 
	      $q2, FPCountStart -> n+1, SimpLevel -> simplev, 
	      NestedExpressions -> False];
    tures = Append[tures, osubb];
    $ol$moms = Alternatives @@ Join[momext, {$q1,$q2}];
  ];
  int1 = int1 /. $d->4;
  If[simplev > 0,
    int1 = Collect[int1, {($q1^2 + _)^_, T[__]}, Simplify]
  ];
  (* introduce replacement symbols for subexpressions that do not depend on
     the loop momenta (to shorten the working expression): *)
  abbc = 0;
  int1 = Collect[int1, {($q1^2 + _)^_, T[__]}, myhold];
  int1 = int1 /. myhold[arg_] :> myhold2[Numerator[arg]]/
		  (denx = Denominator[arg];
		   If[Head[denx] === Times, denx = Select[denx, !NumberQ[#] &]];
		   If[FreeQ[$ol$lsubs, denx], ++abbc; AddSub[denx, abbv[abbc]]];
		   Factor[Denominator[arg]] /. Map[Reverse, $ol$lsubs]);
  int1 = int1 /. myhold2[arg_] :> (++abbc; AddSub[arg,abbv[abbc]]; abbv[abbc]);
  (* subtract UV subloop subtraction term from integral: *)
  int1 = int1 /. ($q1^2 + A_)^(-2) :> (++n; 6*(1-z[n])*z[n]*(A+muUV2)*
		(-2/($q1^2 + A*z[n] - muUV2*(1-z[n]))^3 -
		 (1-2*z[n])*(A+muUV2)/($q1^2 + A*z[n] - muUV2*(1-z[n]))^4));
  (* this step may again leave $q2-dependent terms outside of the T symbols;
     now make sure they are all moved inside the T symbols: *)
  int1 = Collect[int1, {($q1^2 + _)^_, T[__]}, 
		  # /. Power[Dot[arga_,argb_], nA_:1] :> 
			  (T @@ Table[Dot[arga,argb], {nA}]) /; 
			  !FreeQ[Dot[arga,argb], $q2] &];
  int1 = Expand[int1, T[__]];
  int1 = int1 //. {T[arga__] * T[argb__] :> T[arga,argb],
		   T[arga__]^2 :> T[arga,arga]};
  int1 = int1 /. expd_Dot :> QSort2[expd] /. expt_T :> TSort2[expt];
  If[simplev > 1,
    int1 = Collect[int1, {($q1^2 + _)^_, T[__]}, Simplify],
    int1 = Collect[int1, {($q1^2 + _)^_, T[__]}]
  ];
  
  PrintStatus[3, stainf];
  n1 = n;
  (* work on $q2 subloop: *)
  den2 = Union[Select[den, FreeQ[#,$q1]&]];
  exp2 = Map[Count[den, #]&, den2];	(* power of each denominator term *)
  (* introduce Feynman parameters: *)
  den2 = Plus @@ Map[z[++n]*# &, den2];
  sum2 = Plus @@ Table[z[j], {j,n1+1,n}];
  (* variable transformation onto hyper-cube: *)
  trn2 = Table[z[j] -> Product[y[k], {k,n1+1,j}]*If[j===n, 1, 1-y[j+1]], {j,n1+1,n}];
  jac2 = Det[Outer[D, Map[Last,trn2], Table[y[j],{j,n1+1,n}]]];
  trn2 = trn2 /. y[n1+1]->1;
  jac2 = jac2 /. y[n1+1]->1;
  den2 = Collect[DotSimplify[den2], Dot[_,_], Simplify[#, sum2==1]&];
  den2 = den2 /. expd_Dot :> QSort2[expd] /. expt_T :> TSort2[expt] /. rels;
  int2 = den2^(-Plus @@ exp2) * Gamma[Sum[exp2[[j]], {j,n-n1}]] * 
	  (Product[z[j+n1]^(exp2[[j]]-1)/Gamma[exp2[[j]]], {j,n-n1}]);
  (* combine with result from $q1-loop evaluation:
     (the symbol '$tl$placeh' is temporarily introducted to make sure that
      the head of the expression is 'Plus') *)
  int2 = Map[int2 * # &, int1+$tl$placeh]; 
  int2 = int2 /. {($q1^2 + A_)^nA_ :> 1/$tl$dd[$q1,A]^nA /; nA<0,
		  ($q2.$q2 + A_)^nA_ :> 1/$tl$dd[$q2,A]^nA /; nA<0};
  If[simplev > 0,
    int2 = Collect[int2, {$tl$dd[__], T[__]}, Simplify],
    int2 = Collect[int2, {$tl$dd[__], T[__]}]
  ];
  int2 = List @@ Expand[int2, $tl$dd[__]];
  int2 = Select[int2, FreeQ[#,$tl$placeh]&];
  (* introduce one more Feynman parameter to merge the two subloop contrib.: *)
  n++;
  int2 = int2 /. $tl$dd[$q1,A_]^(n1_:1) * $tl$dd[$q2,B_]^(n2_:1) :> 
          Gamma[n1+n2]/Gamma[n1]/Gamma[n2] * z[n]^(n1-1)*(1-z[n])^(n2-1) /
           Collect[z[n]*($q1.$q1+A) + (1-z[n])*($q2.$q2+B), 
	           Dot[_,_], Simplify]^(n1+n2);
  int2 = int2 /. $tl$dd[$q1,A_]^(n1_:1) :> $problem;
  If[!FreeQ[int2, $problem], Print["Whoops! Unexpected error!"]; Abort[]];
  (* shift $q2 to eliminate linear term; but first make some rearrangements
     to make sure that one does not end up with fraction inside the 
     denominator: *)
  int2 = int2 /. (a2_*$q2.$q2 + A_)^nA_ :> (a2^2*$q2.$q2 + a2*A)^nA * a2^(-nA);
  For[i = 1, i <= Length[int2], i++,
    deni = Union[Extract[int2[[i]], Map[Append[#,1]&, 
		    Position[int2[[i]], (_*$q2.$q2 + ___)^(nA_ /; nA<0)]]]];
    If[Length[deni] =!= 1, Print["Error 1"]; Abort[]];
    deni = First[deni] /. Dot->Times;
    k2shift = FullSimplify[Coefficient[deni, $q2, 1]/
    			   (2 Coefficient[deni, $q2, 2])];
    int2[[i]] = int2[[i]] /. $q2 -> $q2 - k2shift;
  ];
  int2 = int2 /. Dot[arga_,argb_] :> Distribute[Dot[arga,argb]] /.
                 T[argt__] :> Distribute[T[argt]];
  (* tensors of odd rank in $q2 are zero: *)
  int2 = int2 /. expt_T :> 0 /; OddQ[Length[Position[expt, $q2]]];
  int2 = int2 //. T[arga___, argp_, argb___] :> 
  		   T[arga, argb]*argp /; FreeQ[argp,$q2];
  int2 = int2 /. T[]->1;
  int2 = Map[Expand[Numerator[#], T[__]]/Denominator[#] &, int2];
  int2 = int2 //. {T[arga__] * T[argb__] :> T[arga,argb],
		   T[arga__]^2 :> T[arga,arga]};
  int2 = int2 /. expd_Dot :> QSort2[expd] /. expt_T :> TSort2[expt];
  int2 = Map[Collect[Numerator[#], T[__]] / 
	      (Denominator[#] /. Power[argp_,expp_] :> Power[
               Collect[DotSimplify[argp], Dot[_,_], Simplify] /. rels, expp]) &, 
	     int2];
  (* tensors of even rank in $q2 evaluate to combinations of metric tensors: *)
  int2 = int2 //. 
        T[argt___, $q2.arga_, $q2.argb_, $q2.argc_, $q2.argd_, $q2.arge_, $q2.argf_] :>
              1/($d*(2+$d)*(4+$d)) * T[argt, $q2.$q2, $q2.$q2, $q2.$q2] * 
		(arga.argb*argc.argd*arge.argf + arga.argc*argb.argd*arge.argf +
		 arga.argd*argb.argc*arge.argf + arga.argb*argc.arge*argd.argf +
		 arga.argc*argb.arge*argd.argf + arga.arge*argb.argc*argd.argf +
		 arga.argb*argc.argf*argd.arge + arga.argc*argb.argf*argd.arge +
		 arga.argf*argb.argc*argd.arge + arga.argd*argb.arge*argc.argf +
		 arga.arge*argb.argd*argc.argf + arga.argd*argb.argf*argc.arge +
		 arga.argf*argb.argd*argc.arge + arga.arge*argb.argf*argc.argd +
		 arga.argf*argb.arge*argc.argd)  /;
		FreeQ[{argt}, $q2.(mom_/;mom=!=$q2)] && 
		FreeQ[{arga,argb,argc,argd,arge,argf}, $q2]  //.
	T[argt___, $q2.arga_, $q2.argb_, $q2.argc_, $q2.argd_] :>
              1/($d*(2+$d)) * T[argt, $q2.$q2, $q2.$q2] * (arga.argb*argc.argd +
	                   arga.argc*argb.argd + arga.argd*argb.argc)  /;
		FreeQ[{argt}, $q2.(mom_/;mom=!=$q2)] && 
		FreeQ[{arga,argb,argc,argd}, $q2]  //.
	T[argt___, $q2.arga_, $q2.argb_] :> 1/$d*T[argt, $q2.$q2] * arga.argb /;
		FreeQ[{argt}, $q2.(mom_/;mom=!=$q2)] && 
		FreeQ[{arga,argb}, $q2];
  int2 = DotSimplify[int2] /. $d->4 /. rels;
  If[simplev > 1,
    int2 = Map[Collect[Numerator[#], T[__], 
		   TurboSimplify[(# //. A_*B_+A_*C_ :> A*(B+C))]&] / 
	    Denominator[#] &, int2]
  ];
  PrintStatus[4, stainf];
  (* introduce replacement symbols for subexpressions that do not depend on
     the loop momenta (to shorten the working expression): *)
  int2 = Map[Collect[Numerator[#], T[__], myhold] / Denominator[#] &, int2];
  int2 = int2 /. myhold[arg_] :> (++abbc; AddSub[arg, abbv[abbc]]; abbv[abbc]);
  (* replace $q1+$q2 integrations by (8-dim.) $q integration: *)
  int2 = Replace[int2, R_*((a1_:1)*$q1.$q1 + (a2_:1)*$q2.$q2 + A_)^nA_ :>
	        	 (a1*a2)^(-2) * ($q^2 + A)^nA *
			 (R /. argt:T[($q2.$q2)..] :> argt/a2^Length[argt]) /; 
		 nA<0, {1}];
  int2 = int2 /. {T[$q2.$q2] -> $q^2*3/6, 
                  T[$q2.$q2, $q2.$q2] -> $q^4*3/10,
        	  T[$q2.$q2, $q2.$q2, $q2.$q2] -> $q^6*3/15,
		  T[$q2.$q2, $q2.$q2, $q2.$q2, $q2.$q2] -> $q^8*3/21,
		  T[$q2.$q2, $q2.$q2, $q2.$q2, $q2.$q2, $q2.$q2] -> $q^10*3/28};
  (* cancel ($q)^2 in numerator against denominator: *)
  int2 = int2 //. ((R1_:1)*$q^na_ + (R2_:0)) * ($q^2 + A_)^nA_ :> 
          R1*($q^2 + A)^(nA+na/2) + 
	  Collect[R2 - R1*(($q^2+A)^(na/2) - $q^na), {$q,Dot[_,_]}]*
	    ($q^2 + A)^nA /; FreeQ[R1, $q];
  If[simplev > 3,
    int2 = Collect[int2, ($q^2 + _)^_, mysimp2],
    int2 = Collect[int2, ($q^2 + _)^_]
  ];
  (* evaluate the loop mom. integration: *)
  int2 = int2 /. ($q^2 + A_)^nA_ :> (-1)^(-nA) * Gamma[-nA-$d]/Gamma[-nA] *
				     (-Simplify[A]-ieps)^($d+nA);
  (* perform variable transformation onto hyper-cube: *)
  int2 = (int2 /. trn1 /. trn2 /. z[nn_] :> y[nn]) * jac1 * jac2;
  $ol$lsubs = ($ol$lsubs /. trn1 /. trn2 /. z[nn_] :> y[nn]);
  If[simplev > 4, 	(* use this with caution!!! *)
    int2 = Collect[int2, (_-ieps)^_, mysimp2] /. (A_-ieps) :> Simplify[A-ieps];
  ];  
  (* rename Feynman parameter variables: *)
  vars = Union[Join[Extract[int2, Position[int2, y[_]]],
  		    Extract[$ol$lsubs, Position[$ol$lsubs, y[_]]]]];
  nvar = Length[vars];
  varr = Table[vars[[j]] -> x[j], {j,nvar}];
  int2 = int2 /. varr;
  $ol$lsubs = $ol$lsubs /. varr;
  (* put all expressions with the same number of Feynman parameters together
     into one term: *)
  For[i = nvar, i > 0, i--,
    res2 = Plus @@ Select[int2, Count[Variables[#], x[_]] === i &];
    If[res2 =!= 0,
      res2 = Append[$ol$lsubs, $ol$res->res2];
      (* remove unused abbreviations: *)
      While[old2 =!= res2,
	old2 = res2;
	res2 = Select[res2, !MatchQ[#, HoldPattern[abbv[_]->_]] || 
    		            !MatchQ[Position[res2, First[#]], {{_,1}}] &];
      ];
      tures = Append[tures, res2]
    ]
  ];
  (* if there is only one expression: *)
  If[MatchQ[Dimensions[tures], {1,_}], tures = First[tures]];
  PrintStatus[5, stainf];
  tures
];
